/*
Copyright (c) 2007 Takahiro Shinozaki

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/* 
   read SSs and output an accumulated SS
   Wed Jun 20 21:09:53 JST 2007
   staka
*/


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "utils.h"
#include "accss.h"



void
pfngets(char *s, int size, FPFNSet *fpfnset)
{
  int p;
  
  for (p=0; p<fpfnset->ninput; p++) {
    fngets(s, size, fpfnset->fpfnin[p]);
  }
}

int
avrmixw(FPFNSet *fpfnset, char *buf)
{
  int nmix;
  float w, ws;
  FPFN *fpfn, *fpfnout;
  int i;
  int p;

  fpfnout = fpfnset->fpfnout;
  if (sscanf(buf, "m%d", &nmix) != 1) {
    fprintf(stderr, "Format error: 'm(int)' is expected.\n");
    exit(1);
  }
  if (MAXMIXPP < nmix+1) {
    fprintf(stderr, "Error: Too large nmix (%d <) %d\n", MAXMIXPP-1, nmix);
    exit(1);
  }
  fprintf(fpfnout->fp, buf);
  for (i=0; i<nmix+1; i++) {
    ws = 0.0;
    for (p=0; p<fpfnset->ninput; p++) {
      fpfn = fpfnset->fpfnin[p];
      fnread(&w, sizeof(float), 1, fpfn);
      SwapFloat32IL(&w);
      ws += w;
    }
    if (accop == AVR) ws /= fpfnset->ninput;
    SwapFloat32IL(&ws);
    fnwrite(&ws, sizeof(float), 1, fpfnout);
  }
  return nmix;
}

int
avrmean(FPFNSet *fpfnset, char *buf)
{
  int dim;
  float m, ms;
  int i;
  FPFN *fpfn, *fpfnout;
  int p;

  fpfnout = fpfnset->fpfnout;
  if (sscanf(buf, "u%d", &dim) != 1) {
    fprintf(stderr, "Format error: 'u(int)' is expected.\n");
    exit(1);
  }
  if (MAXDIMPP < dim+1) {
    fprintf(stderr, "Error: Too large feature dimention (%d <) %d\n", MAXDIMPP-1, dim);
    exit(1);
  }
  fprintf(fpfnout->fp, buf);
  for (i=0; i<dim+1; i++) {
    ms = 0.0;
    for (p=0; p<fpfnset->ninput; p++) {
      fpfn = fpfnset->fpfnin[p];
      fnread(&m, sizeof(float), 1, fpfn);
      SwapFloat32IL(&m);
      ms += m;
    }
    if (accop == AVR) ms /= fpfnset->ninput;
    SwapFloat32IL(&ms);
    fnwrite(&ms, sizeof(float), 1, fpfnout);
  }
  return dim;
}

int
avrdiagvar(FPFNSet *fpfnset, char *buf)
{
  int dim;
  float v, vs;
  int i;
  FPFN *fpfn, *fpfnout;
  int p;

  fpfnout = fpfnset->fpfnout;
  if (buf[0] == 'd') {
    if (sscanf(buf, "d%d", &dim) != 1) {
      fprintf(stderr, "Format error: 'd(int)' is expected.\n");
      exit(1);
    }
  } else {
    if (sscanf(buf, "i%d", &dim) != 1) {
      fprintf(stderr, "Format error: 'i(int)' is expected.\n");
      exit(1);
    }
  }
  if (MAXDIMPP < dim+1) {
    fprintf(stderr, "Error: Too large feature dimention (%d <) %d\n", MAXDIMPP-1, dim);
    exit(1);
  }
  fprintf(fpfnout->fp, buf);
  for (i=0; i<dim+1; i++) {
    vs = 0.0;
    for (p=0; p<fpfnset->ninput; p++) {
      fpfn = fpfnset->fpfnin[p];
      fnread(&v, sizeof(float), 1, fpfn);
      SwapFloat32IL(&v);
      vs += v;
    }
    if (accop == AVR) vs /= fpfnset->ninput;
    SwapFloat32IL(&vs);
    fnwrite(&vs, sizeof(float), 1, fpfnout);
  }
  return dim;
}

int
avrvar(FPFNSet *fpfnset, char *buf)
{
  switch (buf[0]) {
  case 'd':
  case 'i':
    return avrdiagvar(fpfnset, buf);
    break;
  case 'f':
  case 'l':
    printf("Not implemented yet : var f\n");
    exit(1);
    /*    return avrfullvar(fpfn, buf); */
    break;
  default:
    fprintf(stderr, "Format error. 'd','i','f',or,'l' is expected\n");
    exit(1);
  }
  return 0;
}

int
avrtrans(FPFNSet *fpfnset, char *buf)
{
  int nst;
  float t, ts;
  int i, j;
  FPFN *fpfn, *fpfnout;
  int p;

  fpfnout = fpfnset->fpfnout;
  if (sscanf(buf, "t%d", &nst) != 1) {
    fprintf(stderr, "Format error: 't(int)' is expected.\n");
    exit(1);
  }
  if (MAXNSTPP < nst+1) {
    fprintf(stderr, "Error: Too large nst (%d <) %d\n", MAXNSTPP-1, nst);
    exit(1);
  }
  fprintf(fpfnout->fp, buf);
  for (i=0; i<nst+1; i++) {
    for (j=0; j<nst; j++) {
      ts = 0.0;
      for (p=0; p<fpfnset->ninput; p++) {
	fpfn = fpfnset->fpfnin[p];
	fnread(&t, sizeof(float), 1, fpfn);
	SwapFloat32IL(&t);
	ts += t;
      }
      if (accop == AVR) ts /= fpfnset->ninput;
      SwapFloat32IL(&ts);
      fnwrite(&ts, sizeof(float), 1, fpfnout);
    }
  }
  return nst;
}

int
avrheader(FPFNSet *fpfnset, char *buf)
{
  short vecSize, nStream, swidth;
  HSetKind hskind;
  int i;
  FPFN *fpfnout;
  int hk;
  
  fpfnout = fpfnset->fpfnout;
  /* HSetKind */
  pfngets(buf, MAXSTRLEN, fpfnset);
  if (sscanf(buf, "HSetKind=%d", &hk) != 1) {
    fprintf(stderr, "Format error: 'HSetKind(int)' is expected.\n");
    exit(1);
  }
  hskind = (HSetKind)hk;
  fprintf(fpfnout->fp, buf);

  /* VecSize */
  pfngets(buf, MAXSTRLEN, fpfnset);
  if (sscanf(buf, "vecSize=%hd", &vecSize) != 1) {
    fprintf(stderr, "Format error: vecSize is expected.\n");
    exit(1);
  }
  fprintf(fpfnout->fp, buf);

  /* swidth */
  pfngets(buf, MAXSTRLEN, fpfnset);
  if (sscanf(buf, "nStream=%hd", &nStream) != 1) {
    fprintf(stderr, "Format error: nStream is expected.\n");
    exit(1);
  }
  fprintf(fpfnout->fp, buf);
  for (i=0; i<nStream; i++) {
    pfngets(buf, MAXSTRLEN, fpfnset);
    if (sscanf(buf, "%hd", &swidth) != 1) {
      fprintf(stderr, "Format error: swidth is expected.\n");
      exit(1);
    }
    fprintf(fpfnout->fp, buf);
  }

  return hskind;
}

void
version()
{
  printf("Ver 1.0\n");
}

void
usage(char *myname)
{
  version();
  printf("%s [options] ssout ssin(1) ssin(2)...\n", myname);
  printf("Accumulate sufficient statistics files\n");
  printf("  options:\n");
  printf("     -n: average accs\n");
  printf("     -s: sum accs\n");
  printf("     -v n: set verbose level to n\n");
  printf("     -V: print version\n");
}

int
main(int argc, char **argv)
{
  char *ssfile;
  FPFNSet *fpfnset;
  FPFN *fpfnout;
  char buf[MAXSTRLEN];
  HSetKind hskind;
  int numEgs, numEgsTot;
  int nmix;
  int i, j;
  int mark;
  int nstrms;
  int p;
  float totl, totlt;
  int totn, totnt;
  int c;

  while (1) {
    c = getopt(argc, argv, "nsv:V");
    if (c == -1)
      break;
    switch (c) {
    case 'n':
      accop = AVR;
      break;
    case 's':
      accop = SUM;
      break;
    case 'v':
      verbose = atoi(optarg);
      break;
    case 'V':
      version();
      break;
    default:
      break;
    }
  }
  if (argc - optind < 2) {
    usage(argv[0]);
    exit(1);
  }

  ssfile = argv[optind++];
  fpfnout = fnopen(ssfile, "w");
  fpfnset = (FPFNSet *)m_malloc(sizeof(FPFNSet));
  fpfnset->fpfnout = fpfnout;
  fpfnset->ninput = argc-optind;
  fpfnset->fpfnin = (FPFN **)m_malloc(sizeof(FPFN *)*(fpfnset->ninput));
  for (i=optind; i<argc; i++) {
    ssfile = argv[i];
    fpfnset->fpfnin[i-optind] = fnopen(ssfile, "r");
  }

  if (0<verbose) {
    if (accop == AVR) {
      printf("average %d SSs\n", fpfnset->ninput);
    } else {
      printf("sum %d SSs\n", fpfnset->ninput);
    }
  }

  hskind = (HSetKind)avrheader(fpfnset, buf);

  pfngets(buf, MAXSTRLEN, fpfnset);
  while (buf[0] == 'H') { /* hmm */
    fprintf(fpfnout->fp, buf);
    /* hmm name */
    pfngets(buf, MAXSTRLEN, fpfnset);
    fprintf(fpfnout->fp, buf);

    /* num of Examples */
    numEgsTot = 0;
    for (p=0; p<fpfnset->ninput; p++) {
      fnread(&numEgs, sizeof(int), 1, fpfnset->fpfnin[p]);
      SwapInt32IL(&numEgs);
      numEgsTot += numEgs;
    }
    if (accop == AVR) numEgsTot /= fpfnset->ninput;
    SwapInt32IL(&numEgsTot);
    fnwrite(&numEgsTot, sizeof(float), 1, fpfnout);

    pfngets(buf, MAXSTRLEN, fpfnset);
    while (buf[0] == 'S') { /* state */
      fprintf(fpfnout->fp, buf);
      pfngets(buf, MAXSTRLEN, fpfnset);
      while (buf[0] == 'R') { /* stream */
	fprintf(fpfnout->fp, buf);
	pfngets(buf, MAXSTRLEN, fpfnset);
	avrmixw(fpfnset, buf); /* mix weight */
	pfngets(buf, MAXSTRLEN, fpfnset);
	if (buf[0] == 'C') { /* continuous HMM */
	  fprintf(fpfnout->fp, buf);
	  pfngets(buf, MAXSTRLEN, fpfnset);
	  while (buf[0] == 'X') { /* mix components */
	    fprintf(fpfnout->fp, buf);
	    pfngets(buf, MAXSTRLEN, fpfnset);
	    if (buf[0] == 'u') {
	      avrmean(fpfnset, buf);
	      pfngets(buf, MAXSTRLEN, fpfnset);
	    }
	    if (buf[0] == 'd' || buf[0] == 'i' || buf[0] == 'f' || buf[0] == 'l') {
	      avrvar(fpfnset, buf);
	      pfngets(buf, MAXSTRLEN, fpfnset);
	    }
	  } /* mix components */
	  if (buf[0] != 'E') {
	    fprintf(stderr, "Format error: 'E' is expected at the end of mix component definitions.\n");
	  }
	  fprintf(fpfnout->fp, buf);
	} else {
	  /* Non-continuous HMM */
	  /* No definitions here */
	  fprintf(fpfnout->fp, buf);
	}
	pfngets(buf, MAXSTRLEN, fpfnset);
      } /* stream */
      if (buf[0] != 'E') {
	fprintf(stderr, "Format error: 'E' is expected at the end of stream definitions.\n");
      }
      fprintf(fpfnout->fp, buf);
      pfngets(buf, MAXSTRLEN, fpfnset);
    } /* state */
    if (buf[0] != 'E') {
      fprintf(stderr, "Format error: 'E' is expected at the end of state definitions.\n");
    }
    fprintf(fpfnout->fp, buf);
    pfngets(buf, MAXSTRLEN, fpfnset);
    /* transition matrix */
    if (buf[0] == 't') {
      avrtrans(fpfnset, buf);
      pfngets(buf, MAXSTRLEN, fpfnset);
    }
    /* avr marker */
    fprintf(fpfnout->fp, buf); /* M */
    for (p=0; p<fpfnset->ninput; p++) {
      fnread(&mark, sizeof(int), 1, fpfnset->fpfnin[p]); SwapInt32IL(&mark);
      if (mark != 123456) {
	fprintf(stderr, "Format error: Marker 123456 is expected. marker=%d.\n", mark);
	exit(1);
      }
    }
    SwapInt32IL(&mark);
    fnwrite(&mark, sizeof(int), 1, fpfnout);
    pfngets(buf, MAXSTRLEN, fpfnset);
  }
  if (buf[0] != 'E') {
    fprintf(stderr, "Format error: 'E' is expected at the end of HMM definitions.\n");
  }
  fprintf(fpfnout->fp, buf);
  if (hskind == TIEDHS){
    pfngets(buf, MAXSTRLEN, fpfnset);
    if (sscanf(buf, "s%d", &nstrms) != 1) {
      fprintf(stderr, "Format error: 's(int)' is expected.\n");
      exit(1);
    }
    fprintf(fpfnout->fp, buf);
    for (i=0; i<nstrms; i++) {
      pfngets(buf, MAXSTRLEN, fpfnset);
      if (sscanf(buf, "m%d", &nmix) != 1) {
	fprintf(stderr, "Format error: 'm(int)' is expected.\n");
	exit(1);
      }
      fprintf(fpfnout->fp, buf);
      for (j=0; j<nmix; j++) {
	pfngets(buf, MAXSTRLEN, fpfnset);
	avrmean(fpfnset, buf);
	pfngets(buf, MAXSTRLEN, fpfnset);
	avrvar(fpfnset, buf);
      }
    }
  }
  pfngets(buf, MAXSTRLEN, fpfnset);
  if (buf[0] != 'E') {
    fprintf(stderr, "Format error: 'E' is expected at the end of HMM definitions.\n");
  }
  fprintf(fpfnout->fp, buf);

  totlt = 0.0;
  for (p=0; p<fpfnset->ninput; p++) {
    fnread(&totl, sizeof(float), 1, fpfnset->fpfnin[p]);
    SwapFloat32IL(&totl);
    totlt += totl;
  }
  if (accop == AVR) totlt /= fpfnset->ninput;
  SwapFloat32IL(&totlt);
  fnwrite(&totlt, sizeof(float), 1, fpfnout);

  totnt = 0;
  for (p=0; p<fpfnset->ninput; p++) {
    fnread(&totn, sizeof(int), 1, fpfnset->fpfnin[p]);
    SwapInt32IL(&totn);
    totnt += totn;
  }
  if (accop == AVR) totnt /= fpfnset->ninput;
  SwapInt32IL(&totnt);
  fnwrite(&totnt, sizeof(int), 1, fpfnout);

  fnclose(fpfnout);
  return 0;
}
