/*
Copyright (c) 2007 Takahiro Shinozaki

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

/* 
   dumpss
   dump ss file
   staka
   Thu Jun 21 19:50:16 JST 2007
*/

#include <stdio.h>
#include "utils.h"
#include "dumpss.h"

#define MAXSTRLEN 256
#define MAXMIXPP 4096
#define MAXDIMPP 1024
#define MAXNSTPP 64

enum _HSetKind {PLAINHS, SHAREDHS, TIEDHS, DISCRETEHS};
typedef enum _HSetKind HSetKind;


int
dumpmixw(FPFN *fpfn, char *buf)
{
  int nmix;
  float mixw[MAXMIXPP], w;
  int i;

  if (sscanf(buf, "m%d", &nmix) != 1) {
    fprintf(stderr, "Format error: 'm(int)' is expected. %s\n", fpfn->path);
    exit(1);
  }
  if (MAXMIXPP < nmix+1) {
    fprintf(stderr, "Error: Too large nmix (%d <) %d\n", MAXMIXPP-1, nmix);
    exit(1);
  }
  fnread(mixw, sizeof(float), nmix+1, fpfn);
  for (i=0; i<nmix+1; i++) SwapInt32IL((int *)&(mixw[i]));
  printf("mixw %d\n", nmix);
  for (i=0; i<nmix; i++) {
    w=mixw[i];
    printf("%f ", w);
  }
  w=mixw[nmix];
  printf(": %f\n", w);
  return nmix;
}

int
dumpmean(FPFN *fpfn, char *buf)
{
  int dim;
  float mean[MAXDIMPP], m;
  int i;

  if (sscanf(buf, "u%d", &dim) != 1) {
    fprintf(stderr, "Format error: 'u(int)' is expected. %s\n", fpfn->path);
    exit(1);
  }
  if (MAXDIMPP < dim+1) {
    fprintf(stderr, "Error: Too large feature dimention (%d <) %d\n", MAXDIMPP-1, dim);
    exit(1);
  }
  fnread(mean, sizeof(float), dim+1, fpfn);
  for (i=0; i<dim+1; i++) SwapInt32IL((int *)&(mean[i]));
  printf("mean %d\n", dim);
  for (i=0; i<dim; i++) {
    m=mean[i];
    printf("%f ", m);
  }
  m=mean[dim];
  printf(": %f\n", m);
  return dim;
}

int
dumpdiagvar(FPFN *fpfn, char *buf)
{
  int dim;
  float var[MAXDIMPP], v;
  int i;

  if (buf[0] == 'd') {
    if (sscanf(buf, "d%d", &dim) != 1) {
      fprintf(stderr, "Format error: 'd(int)' is expected. %s\n", fpfn->path);
      exit(1);
    }
  } else {
    if (sscanf(buf, "i%d", &dim) != 1) {
      fprintf(stderr, "Format error: 'i(int)' is expected. %s\n", fpfn->path);
      exit(1);
    }
  }
  if (MAXDIMPP < dim+1) {
    fprintf(stderr, "Error: Too large feature dimention (%d <) %d\n", MAXDIMPP-1, dim);
    exit(1);
  }
  fnread(var, sizeof(float), dim+1, fpfn);
  for (i=0; i<dim+1; i++) SwapInt32IL((int *)&(var[i]));
  printf("var %d\n", dim);
  for (i=0; i<dim; i++) {
    v=var[i];
    printf("%f ", v);
  }
  v=var[dim];
  printf(": %f\n", v);
  return dim;
}

int
dumpvar(FPFN *fpfn, char *buf)
{
  switch (buf[0]) {
  case 'd':
  case 'i':
    return dumpdiagvar(fpfn, buf);
    break;
  case 'f':
  case 'l':
    printf("Not implemented yet : var f\n");
    exit(1);
    /*    return dumpfullvar(fpfn, buf); */
    break;
  default:
    fprintf(stderr, "Format error. 'd','i','f',or,'l' is expected\n");
    exit(1);
  }
  return 0;
}

int
dumptrans(FPFN *fpfn, char *buf)
{
  int nst;
  float trp[(MAXNSTPP-1)*MAXNSTPP], p;
  int i, j;

  if (sscanf(buf, "t%d", &nst) != 1) {
    fprintf(stderr, "Format error: 't(int)' is expected. %s\n", fpfn->path);
    exit(1);
  }
  if (MAXNSTPP < nst+1) {
    fprintf(stderr, "Error: Too large nst (%d <) %d\n", MAXNSTPP-1, nst);
    exit(1);
  }
  fnread(trp, sizeof(float), nst*(nst+1), fpfn);
  for (i=0; i<nst*(nst+1); i++) SwapInt32IL((int *)&(trp[i]));
  printf("trp %d\n", nst);
  for (i=0; i<nst+1; i++) {
    for (j=0; j<nst; j++) {
      p=trp[i*nst+j];
      printf("%f ", p);
    }
    printf("\n");
  }
  return nst;
}

int
dumpheader(FPFN *fpfn, char *buf)
{
  short vecSize, nStream, swidth;
  HSetKind hskind;
  int i;
  int hk;
  

  /* HSetKind */
  fngets(buf, MAXSTRLEN, fpfn);
  if (sscanf(buf, "HSetKind=%d", &hk) != 1) {
    fprintf(stderr, "Format error: 'HSetKind(int)' is expected. %s\n", fpfn->path);
    exit(1);
  }
  hskind = (HSetKind)hk;
  switch (hskind) {
  case PLAINHS:
    printf("HSetKind: PLAINHS\n");
    break;
  case SHAREDHS:
    printf("HSetKind: SHAREDHS\n");
    break;
  case TIEDHS:
    printf("HSetKind: TIEDHS\n");
    break;
  case DISCRETEHS:
    printf("HSetKind: DISCRETEHS\n");
    break;
  default:
    break;
  }

  /* VecSize */
  fngets(buf, MAXSTRLEN, fpfn);
  if (sscanf(buf, "vecSize=%hd", &vecSize) != 1) {
    fprintf(stderr, "Format error: vecSize is expected. %s\n", fpfn->path);
    exit(1);
  }
  printf("vecSize = %hd\n", vecSize);

  /* swidth */
  fngets(buf, MAXSTRLEN, fpfn);
  if (sscanf(buf, "nStream=%hd", &nStream) != 1) {
    fprintf(stderr, "Format error: nStream is expected. %s\n", fpfn->path);
    exit(1);
  }
  printf("nStream = %hd :", nStream);
  for (i=0; i<nStream; i++) {
    fngets(buf, MAXSTRLEN, fpfn);
    if (sscanf(buf, "%hd", &swidth) != 1) {
      fprintf(stderr, "Format error: swidth is expected. %s\n", fpfn->path);
      exit(1);
    }
    printf(" %hd", swidth);
  }
  printf("\n");

  return hskind;
}

int
main(int argc, char **argv)
{
  char *ssfile;
  FPFN *fpfn;
  char buf[MAXSTRLEN];
  HSetKind hskind;
  int numEgs;
  int nmix;
  int i, j;
  int mark;
  int nstrms;
  float totl;
  int totn;

  if (argc != 2) {
    fprintf(stderr, "%s ss\n", argv[0]);
    exit(1);
  }
  ssfile = argv[1];
  fpfn = fnopen(ssfile, "r");

  hskind = (HSetKind)dumpheader(fpfn, buf);

  fngets(buf, MAXSTRLEN, fpfn);
  while (buf[0] == 'H') { /* hmm */
    /* hmm name */
    fngets(buf, MAXSTRLEN, fpfn);
    printf("%s", buf);

    /* num of Examples */
    fnread(&numEgs, sizeof(int), 1, fpfn); SwapInt32IL(&numEgs);
    printf("numEgs %d\n", numEgs);

    fngets(buf, MAXSTRLEN, fpfn);
    while (buf[0] == 'S') { /* state */
      fngets(buf, MAXSTRLEN, fpfn);
      while (buf[0] == 'R') { /* stream */
	fngets(buf, MAXSTRLEN, fpfn);
	nmix = dumpmixw(fpfn, buf); /* mix weight */
	fngets(buf, MAXSTRLEN, fpfn);
	if (buf[0] == 'C') { /* continuous HMM */
	  fngets(buf, MAXSTRLEN, fpfn);
	  while (buf[0] == 'X') { /* mix components */
	    fngets(buf, MAXSTRLEN, fpfn);
	    if (buf[0] == 'u') {
	      dumpmean(fpfn, buf);
	      fngets(buf, MAXSTRLEN, fpfn);
	    }
	    if (buf[0] == 'd' || buf[0] == 'i' || buf[0] == 'f' || buf[0] == 'l') {
	      dumpvar(fpfn, buf);
	      fngets(buf, MAXSTRLEN, fpfn);
	    }
	  } /* mix components */
	  if (buf[0] != 'E') {
	    fprintf(stderr, "Format error: 'E' is expected at the end of mix component definitions.\n");
	  }
	} else {
	  /* Non-continuous HMM */
	  /* No definitions here */
	}
	fngets(buf, MAXSTRLEN, fpfn);
      } /* stream */
      if (buf[0] != 'E') {
	fprintf(stderr, "Format error: 'E' is expected at the end of stream definitions.\n");
      }
      fngets(buf, MAXSTRLEN, fpfn);
    } /* state */
    if (buf[0] != 'E') {
      fprintf(stderr, "Format error: 'E' is expected at the end of state definitions.\n");
    }
    fngets(buf, MAXSTRLEN, fpfn);
    /* transition matrix */
    if (buf[0] == 't') {
      dumptrans(fpfn, buf);
      fngets(buf, MAXSTRLEN, fpfn);
    }
    /* dump marker */
    fnread(&mark, sizeof(int), 1, fpfn); SwapInt32IL(&mark);
    if (mark != 123456) {
      fprintf(stderr, "Format error: Marker 123456 is expected. marker=%d.\n", mark);
      exit(1);
    }
    fngets(buf, MAXSTRLEN, fpfn);
  }
  if (buf[0] != 'E') {
    fprintf(stderr, "Format error: 'E' is expected at the end of HMM definitions.\n");
  }
  if (hskind == TIEDHS){
    fngets(buf, MAXSTRLEN, fpfn);
    if (sscanf(buf, "s%d", &nstrms) != 1) {
      fprintf(stderr, "Format error: 's(int)' is expected. %s\n", fpfn->path);
      exit(1);
    }
    for (i=0; i<nstrms; i++) {
      fngets(buf, MAXSTRLEN, fpfn);
      if (sscanf(buf, "m%d", &nmix) != 1) {
	fprintf(stderr, "Format error: 'm(int)' is expected. %s\n", fpfn->path);
	exit(1);
      }
      for (j=0; j<nmix; j++) {
	fngets(buf, MAXSTRLEN, fpfn);
	dumpmean(fpfn, buf);
	fngets(buf, MAXSTRLEN, fpfn);
	dumpvar(fpfn, buf);
      }
    }
  }

  fngets(buf, MAXSTRLEN, fpfn); /* E */

  fnread(&totl, sizeof(float), 1, fpfn);
  SwapFloat32IL(&(totl));
  printf("totl = %f\n", totl);

  fnread(&totn, sizeof(int), 1, fpfn);
  SwapInt32IL(&totn);
  printf("totn = %d\n", totn);

  return 0;
}
