/* utils.c
 * Wed Sep 28 17:41:45 PDT 2005
 * staka
 *
 * utility functions
 */

#include "utils.h"

void *
m_malloc(size_t size)
{

#ifdef CHECKMALLOC
  return check_malloc(size);
#else
  void *pt;

  if ((pt=malloc(size))==NULL) {
    perror("m_malloc");
    exit(1);
  }
  return pt;
#endif
}

void *
m_realloc(void *ptr, size_t size)
{
#ifdef CHECKMALLOC
  return check_realloc(ptr, size);
#else
  void *pt;

  if ((pt=realloc(ptr, size))==NULL) {
    perror("m_realloc");
    exit(1);
  }
  return pt;
#endif
}


void
m_free(void *ptr)
{
#ifdef CHECKMALLOC
  check_free(ptr);
  return;
#else 
  free(ptr);
#endif
}

int
isLittle(void)
{
  int n = 1;
  if (*(char*)&n) {
    return(1);
  } else {
    return(0);
  }
}

/* apply byte swap if this machine uses little endian */
void SwapInt32IL(int *p)
{
   char temp,*q;
   
   if (isLittle()) {
     q = (char*) p;
     temp = *q; *q = *(q+3); *(q+3) = temp;
     temp = *(q+1); *(q+1) = *(q+2); *(q+2) = temp;
   }
}

/* apply byte swap if this machine uses little endian */
void SwapFloat32IL(float *p)
{
   char temp,*q;
   
   if (isLittle()) {
     q = (char*) p;
     temp = *q; *q = *(q+3); *(q+3) = temp;
     temp = *(q+1); *(q+1) = *(q+2); *(q+2) = temp;
   }
}

char *
m_strdup(const char *s)
{
  char *cp;
  if ((cp=strdup(s))==NULL) {
    perror(s);
    exit(1);
  } else {
    return cp;
  }
}

char *
m_chomp(char *str)
{
  int len;
  
  len = strlen(str);
  if (len>0) {
    if (str[len-1] == '\n') {
      str[len-1] = '\0';
    }
  }
  return str;
}

FILE *
m_fopen(const char *path, const char *mode)
{
  FILE *fp;
  if ((fp=fopen(path, mode))==NULL) {
    perror(path);
    exit(1);
  }
  return fp;
}

FPFN *
fnopen(const char *path, const char *mode)
{
  FPFN *fpfn;
  fpfn = (FPFN *)m_malloc(sizeof(FPFN));
  if ((fpfn->fp=fopen(path, mode))==NULL) {
    perror(path);
    exit(1);
  }
  fpfn->path = m_strdup(path);
  return fpfn;
}

int
m_fclose(FILE *stream)
{
  if (fclose(stream)!=0) {
    perror("fclose");
    exit(1);
  }
  return 0;
}

int
fnclose(FPFN *fpfn)
{
  if (fclose(fpfn->fp)!=0) {
    perror(fpfn->path);
    exit(1);
  }
  m_free(fpfn->path);
  m_free(fpfn);
  return 0;
}

size_t
m_fread(void *ptr, size_t size, size_t nmemb, FILE *stream, char *fname)
{
  size_t retv;
  char buf[1000];

  retv = fread(ptr, size, nmemb, stream);
  if (retv == nmemb) {
    return(nmemb);
  } else {
    sprintf(buf, "Error read failed %s: retv %d size %d nmemb %d", fname, (int)retv, (int)size, (int)nmemb);
    perror(buf);
    exit(1);
  }
}

size_t
fnread(void *ptr, size_t size, size_t nmemb, FPFN *fpfn)
{
  size_t retv;
  char buf[1000];

  retv = fread(ptr, size, nmemb, fpfn->fp);
  if (retv == nmemb) {
    return(nmemb);
  } else {
    sprintf(buf, "Error read failed %s: retv %d size %d nmemb %d", fpfn->path, (int)retv, (int)size, (int)nmemb);
    perror(buf);
    exit(1);
  }
}

size_t
fnwrite(const void *ptr, size_t size, size_t nmemb, FPFN *fpfn)
{
  size_t retv;
  char buf[1000];

  retv = fwrite(ptr, size, nmemb, fpfn->fp);
  if (retv == nmemb) {
    return(nmemb);
  } else {
    sprintf(buf, "Error read failed %s: retv %d size %d nmemb %d", fpfn->path, (int)retv, (int)size, (int)nmemb);
    perror(buf);
    exit(1);
  }
}

void
m_fgets(char *s, int size, FILE *stream)
{
  if (fgets(s, size, stream) == NULL) {
    if (feof(stream)) {
      fprintf(stderr, "Can't read stream: EOF\n");
    } else {
      fprintf(stderr, "Can't read stream.\n");
      perror("m_fgets");
    }
    exit(1);
  }
}

void
fngets(char *s, int size, FPFN *fpfn)
{
  if (fgets(s, size, fpfn->fp) == NULL) {
    if (feof(fpfn->fp)) {
      fprintf(stderr, "Can't read stream: EOF %s\n", fpfn->path);
    } else {
      fprintf(stderr, "Can't read stream. %s\n", fpfn->path);
      perror("fngets");
    }
    exit(1);
  }
}

/* File access control using fcntl. (Advisory file locking)
 * The file must be writable. I.e. it 
 * must be opend with r+, w, w+, a or a+)
 */
int
m_lock_fcntl(FILE *fp)
{
  struct flock lock;
  int fd;

  lock.l_type = F_WRLCK;	/* Exclusive lock (This requires the file is writable) */
  lock.l_whence = SEEK_SET;
  lock.l_start = 0;
  lock.l_len = 0;
  lock.l_pid = getpid();

  fd = fileno(fp);
  if (fcntl(fd, F_SETLKW, &lock)) {
    perror("fcntl");
    exit(1);
  }
  return fd;
}

/* Unlock the file.
 * Even without the call of this fuction, the file is 
 * automatically unlocked when it is closed or the process terminate.
 */
int
m_unlock_fcntl(FILE *fp)
{
  struct flock lock;
  int fd;

  lock.l_type = F_WRLCK;
  lock.l_whence = SEEK_SET;
  lock.l_start = 0;
  lock.l_len = 0;
  lock.l_pid = getpid();

  fd = fileno(fp);
  if (fcntl(fd, F_UNLCK, &lock)) {
    perror("fcntl");
    exit(1);
  }
  return fd;
}
